@@ -9,7 +9,9 @@ import Control.Monad.Fail (MonadFail)
 import Control.Monad.Reader
 import Data.IORef
 import Database.Redis.Protocol
+import Control.Monad.IO.Unlift (MonadUnliftIO)
 import qualified Database.Redis.ProtocolPipelining as PP
+import qualified Database.Redis.Cluster as Cluster
 
 -- |Context for normal command execution, outside of transactions. Use
 --  'runRedis' to run actions of this type.
@@ -18,12 +20,13 @@ import qualified Database.Redis.ProtocolPipelining as PP
 --  possibility of Redis returning an 'Error' reply.
 newtype Redis a =
   Redis (ReaderT RedisEnv IO a)
-  deriving (Monad, MonadIO, Functor, Applicative)
+  deriving (Monad, MonadIO, Functor, Applicative, MonadUnliftIO)
 #if __GLASGOW_HASKELL__ > 711
 deriving instance MonadFail Redis
 #endif
-data RedisEnv =
-  Env
-    { envConn :: PP.Connection
-    , envLastReply :: IORef Reply
-    }
+data RedisEnv
+    = NonClusteredEnv { envConn :: PP.Connection, envLastReply :: IORef Reply }
+    | ClusteredEnv
+        { refreshAction :: IO Cluster.ShardMap
+        , connection :: Cluster.Connection
+        }
